#!/usr/bin/perl

sub promptUser;
sub trim;
sub CheckForObjFile_Generic;
sub GenerateObjectFileList;

sub strlist_union;          # addition
sub strlist_complement;     # difference
sub strlist_intersection;   # retrieve common elements


# =====================  EDIT TO SUIT ========================= #

#   Location and library name for LAPACK

$LAPACK_DIR      = '/home/dmay/software/lapack-3.1.1';
$LAPACK_LIBNAME  = 'liblapack_LINUX_x86_goto.a';

#   Location and library name for GotoBLAS
$GOTO_DIR      = '/home/dmay/software/GotoBLAS';
$GOTO_LIBNAME  = 'libgoto_opteron-r1.21.a';

# ============================================================= #


$LAPACK_LIB = $LAPACK_DIR . "/" . $LAPACK_LIBNAME;
$GOTO_LIB = $GOTO_DIR . "/" . $GOTO_LIBNAME;

## Merged lib names
$GOTOBLAS_LIBNAME   = 'libgoto_blas.a';
$GOTOLAPACK_LIBNAME = 'libgoto_lapack.a';

print "Extracting all object files from GotoBLAS:--------  $GOTO_DIR\/$GOTO_LIBNAME  ";
@goto_ofiles   = GenerateObjectFileList( $GOTO_LIB );
$N = @goto_ofiles;
print ",    Found $N symbols \n";

print "Extracting all object files from LAPACK:----------  $LAPACK_DIR\/$LAPACK_LIBNAME  ";
@lapack_ofiles = GenerateObjectFileList( $LAPACK_LIB );
$N = @lapack_ofiles;
print ",    Found $N symbols \n";

#print "================================================================================================\n";
#print "Symbols from $GOTO_LIB:\n    { @goto_ofiles } \n\n";
#print "================================================================================================\n";
#print "Symbols from $LAPACK_LIB:\n    { @lapack_ofiles } \n\n";


# Compare symbol list and determine the duplicates
print "Determining duplicate symbols:--------------------  ";
@duplicate_symbols = strlist_intersection( \@goto_ofiles, \@lapack_ofiles );
$N = @duplicate_symbols;
print ",    Found $N duplicate symbols \n";

#  Remove duplicates from lapack list, these will be 
#  replaced by the lapack symbols from the GotoBLAS library

print "Isolated BLAS symbols:----------------------------  ";
@goto_ofiles_BLAS = strlist_complement( \@goto_ofiles, \@duplicate_symbols );
$N = @goto_ofiles_BLAS;
print ",    Found $N unique blas symbols \n";

@goto_ofiles_LAPACK = @duplicate_symbols;

print "Isolated LAPACK symbols:--------------------------  ";
@lapack_ofiles_NONGOTO = strlist_complement( \@lapack_ofiles, \@duplicate_symbols );
$N = @lapack_ofiles_NONGOTO;
print ",    Found $N unique lapack symbols \n";

#print "================================================================================================\n";
#print "Duplicate symbols:\n    { @duplicate_symbols } \n\n";
#print "================================================================================================\n";
#print "BLAS symbols:\n    { @goto_ofiles_BLAS } \n\n";
#print "================================================================================================\n";
#print "LAPACK symbols:\n    { @lapack_ofiles_NONGOTO } \n\n";


# Rename all symbols which originate from GotoBLAS
$blas_sym_dir   = 'GotoBLAS_symbols';
$lapack_sym_dir = 'GotoLAPACK_symbols';

`rm -f *.o`;


#### BLAS SYMBOLS ####

$dir = $blas_sym_dir;
unless(-d $dir){
	$rebuild = 'true';
}
else {
	$ans = &promptUser("Directory $dir exists. Do you wish to rebuild ALL object files? (y/n)", "y");
	if( $ans eq 'y' ) {
		$rebuild = 'true';
	}
	else {
		$rebuild = 'false';
	}
}

if( $rebuild eq 'true' ) {
	`rm -rf $dir`; 
	`mkdir $dir`;
	print "Processing GotoBLAS, BLAS symbols: \n";
	$N = @goto_ofiles_BLAS;
	$ten_percent = $N / 10.0;
	$perc_done = 0.0;

	for( $i=0; $i<@goto_ofiles_BLAS; $i++ ) {
	    $obj = @goto_ofiles_BLAS->[$i];
            $inc = $i+1;
#	    print "    $obj ($inc/$N) \n";
	    if( $i%$ten_percent == 0 ) { print "    Done $perc_done\% - [$i/$N] \n"; $perc_done = $perc_done + 10 }
	    `ar -xo $GOTO_LIB $obj`;
	    `mv $obj $dir/goto_$obj`;
	}
}


#### LAPACK SYMBOLS ####
$dir = $lapack_sym_dir;

unless(-d $dir){
	$rebuild = 'true';
}
else {
	$ans = &promptUser("Directory $dir exists. Do you wish to rebuild ALL object files? (y/n)", "y");
	if( $ans eq 'y' ) {
		$rebuild = 'true';
	}
	else {
		$rebuild = 'false';
	}
}

if( $rebuild eq 'true' ) {

	`rm -rf $dir`;
	`mkdir $dir`;
	# Get the goto-lapack symbols
	print "Processing GotoBLAS, LAPACK symbols: \n";
	print "    ";     # leave blank space for the .... progress indicator
	$N = @goto_ofiles_LAPACK;

	for( $i=0; $i<@goto_ofiles_LAPACK; $i++ ) {
		$obj = @goto_ofiles_LAPACK->[$i];
		print ".";
		`ar -xo $GOTO_LIB $obj`;
		`mv $obj $dir/goto_$obj`;
	}
	print "    Finished \n";

	# Get the non duplicate lapack symbols, do not rename the object files
	print "Processing LAPACK symbols: \n";
	$N = @lapack_ofiles_NONGOTO;
	$ten_percent = $N / 10.0;
	$perc_done = 0.0;

	for( $i=0; $i<@lapack_ofiles_NONGOTO; $i++ ) {
		$obj = @lapack_ofiles_NONGOTO->[$i];
#		print "    $obj ($i/$N) \n";
		if( $i%$ten_percent == 0 ) { print "    Done $perc_done\% - [$i/$N] \n"; $perc_done = $perc_done + 10 }
		`ar -xo $LAPACK_LIB $obj`;
		`mv $obj $dir/$obj`;
	}
}
## Now create the libraries
print "Creating library:---------------------------------  $GOTOBLAS_LIBNAME  \n";
`ar rcs $GOTOBLAS_LIBNAME $blas_sym_dir/*.o`;

print "Creating library:---------------------------------  $GOTOLAPACK_LIBNAME  \n";
`ar rcs $GOTOLAPACK_LIBNAME $lapack_sym_dir/*.o`;



# <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< | >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> #


sub GenerateObjectFileList
{
    my $lib_file = $_[0];
    my @ofile_list;
    my $add = 0;
    
    # Run nm over file and send output into file
    my $nm_log = "nm_output.txt";
    `nm $lib_file > $nm_log`;

    # Scan output file and look for a particular pattern to extract *.o files
    open( DAT, $nm_log ) || die( "Could not open file ($nm_log) ! \n" );

    my @raw_data=<DAT>;

    foreach $line (@raw_data) {
        chomp( $line );
        my $tline = trim( $line );
    
        $found_obj = CheckForObjFile_Generic( $tline, $obj_file );
        if( $found_obj eq 'true' ) {
#            print "    Found object file, $obj_file \n";
            $ofile_list[$add] = $obj_file;
            $add++;
        }
    }
    close( DAT );
    
    return @ofile_list;    
}

## =============================================================================
##  Perl trim function to remove whitespace from the start and end of the string
##  <input>   Arg1: string to be trimmed
##  <returnt> Arg1: the trimmed string
sub trim($)
{
    my $string = shift;
    $string =~ s/^\s+//;
    $string =~ s/\s+$//;
    return $string;
}


##
##  Look for the pattern
##      /usr/local/lib/libpng.a(pngpread.o):
##      $lib_file(*.o):
sub CheckForObjFile_OSX
{
    $lib = $_[0];
    $str = $_[1];

    

}

##
##  Look for the pattern
##      *.o
sub CheckForObjFile_Generic
{
    my $str = $_[0];
    my $found;

    if( $str =~ m/.*\.o/ ){
#        print "$str --  ";   
#        print "$& --  ";
        
        # remove all none character which is not a word or number from the left of 
        # the matched string, i.e. $&
        my $match = $&;
        my $L = length( $match );
        $tmp = $match;
        
        $k = $L;
        while( $k != 0 ) {
            $a = chop( $tmp );
            if( $a eq "" ) {   $offset = $k; last;  }
            elsif( $a eq "/" ) {  $offset = $k; last; }
            elsif( $a eq "(" ) {  $offset = $k; last; }
            elsif( $a eq ")" ) {  $offset = $k; last; }

            $k = $k - 1;
        }

        my $fragment =  substr $match, $offset;
#        print "    Found object file, $fragment \n";
        $_[1] = $fragment;
 
       $found = 'true'; 
    }
    else {
        $found = 'false';
    }

    return $found;
}


# ===========================
#  return A \cup B
#
sub strlist_union 
{
    my ($A, $B) = @_;
	my @C;

#	print "A = @$A  \n";
#	print "B = @$B  \n";

	for( $i=0; $i<@$A; $i++ ) {
#		print "A = $A->[$i] \n";
		$C[ $i ] = $A->[ $i ];
	}
	for( $i=0; $i<@$B; $i++ ) {
#		print "B = $B->[$i] \n";
		$C[ $i+@$A ] = $B->[ $i ];
	}

	return @C;
}

# ===========================
#  return A \ B
#
sub strlist_complement
{
    my ($A, $B) = @_;
	my @C;
    my $add = 0;

    # if any part of B is in A, then do not add into A
    for( $i=0; $i<@$A; $i++ ) {
        $tmpA = $A->[ $i ];        
        $add_Ai = 'true';
        for( $j=0; $j<@$B; $j++ ) {
            $tmpB = $B->[ $j ];
    
            if( $tmpA eq $tmpB ) {
                $add_Ai = 'false';
                last;
             }
        }
        
        if( $add_Ai eq 'true' ) {
            $C[ $add ] =  $tmpA;
            $add++;
        }
        
	}

	return @C;
}



# ===========================
#  return A \cap B
#
sub strlist_intersection
{
    my ($A, $B) = @_;
	my @C;
    my $add = 0;

    for( $i=0; $i<@$A; $i++ ) {
        $tmpA = $A->[ $i ];        
        $add_Ai = 'false';
        for( $j=0; $j<@$B; $j++ ) {
            $tmpB = $B->[ $j ];
    
            if( $tmpA eq $tmpB ) {
                $add_Ai = 'true';
                last;
             }
        }
        
        if( $add_Ai eq 'true' ) {
            $C[ $add ] =  $tmpA;
            $add++;
        }
        
	}

    return @C;
}



#----------------------------(  promptUser  )-----------------------------#
#                                                                         #
#  FUNCTION:	promptUser                                                #
#                                                                         #
#  PURPOSE:	Prompt the user for some type of input, and return the    #
#		input back to the calling program.                        #
#                                                                         #
#  ARGS:	$promptString - what you want to prompt the user with     #
#		$defaultValue - (optional) a default value for the prompt #
#                                                                         #
#-------------------------------------------------------------------------#

sub promptUser {

   #-------------------------------------------------------------------#
   #  two possible input arguments - $promptString, and $defaultValue  #
   #  make the input arguments local variables.                        #
   #-------------------------------------------------------------------#

   local($promptString,$defaultValue) = @_;

   #-------------------------------------------------------------------#
   #  if there is a default value, use the first print statement; if   #
   #  no default is provided, print the second string.                 #
   #-------------------------------------------------------------------#

   if ($defaultValue) {
      print $promptString, "[", $defaultValue, "]: ";
   } else {
      print $promptString, ": ";
   }

   $| = 1;               # force a flush after our print
   $_ = <STDIN>;         # get the input from STDIN (presumably the keyboard)


   #------------------------------------------------------------------#
   # remove the newline character from the end of the input the user  #
   # gave us.                                                         #
   #------------------------------------------------------------------#

   chomp;

   #-----------------------------------------------------------------#
   #  if we had a $default value, and the user gave us input, then   #
   #  return the input; if we had a default, and they gave us no     #
   #  no input, return the $defaultValue.                            #
   #                                                                 # 
   #  if we did not have a default value, then just return whatever  #
   #  the user gave us.  if they just hit the <enter> key,           #
   #  the calling routine will have to deal with that.               #
   #-----------------------------------------------------------------#

   if ("$defaultValue") {
      return $_ ? $_ : $defaultValue;    # return $_ if it has a value
   } else {
      return $_;
   }
}

